# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from mailbox import Maildir, mbox

from .conftest import MailFactory, check_box, skip_windows

""" Some invalid dates are not supported on Windows as they produce negative
timestamps. See:
* https://github.com/arrow-py/arrow/issues/675
* https://github.com/arrow-py/arrow/pull/745
"""

invalid_date_mail_1 = MailFactory(date_rfc2822="Thu, 13 Dec 101 15:30 WET")
invalid_date_mail_2 = MailFactory(date_rfc2822="Thu, 13 Dec 102 15:30 WET")


@skip_windows
def test_invalid_date_parsing_noop(invoke, make_box):
    """Mails with strange non-standard dates gets parsed anyway and
    grouped into duplicate sets. No deduplication happen: mails groups shares
    the same metadata."""
    box_path, box_type = make_box(
        Maildir,
        [
            invalid_date_mail_1,
            invalid_date_mail_2,
            invalid_date_mail_2,
            invalid_date_mail_1,
            invalid_date_mail_1,
        ],
    )

    result = invoke("--strategy=select-newest", "--action=delete-selected", box_path)

    assert result.exit_code == 0

    check_box(
        box_path,
        box_type,
        content=[
            invalid_date_mail_1,
            invalid_date_mail_1,
            invalid_date_mail_1,
            invalid_date_mail_2,
            invalid_date_mail_2,
        ],
    )


@skip_windows
def test_invalid_date_parsing_dedup(invoke, make_box):
    """Mails with strange non-standard dates gets parsed anyway and
    deduplicated if we reduce the source of hashed headers."""
    box_path, box_type = make_box(
        Maildir,
        [
            invalid_date_mail_1,
            invalid_date_mail_2,
            invalid_date_mail_2,
            invalid_date_mail_1,
            invalid_date_mail_1,
        ],
    )

    result = invoke(
        "--hash-header=message-id",
        "--hash-header=from",
        "--hash-header=to",
        "--hash-header=subject",
        "--strategy=select-newest",
        "--action=delete-selected",
        box_path,
    )

    assert result.exit_code == 0

    check_box(
        box_path,
        box_type,
        content=[
            invalid_date_mail_1,
            invalid_date_mail_1,
            invalid_date_mail_1,
        ],
    )



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Expose package-wide elements. """

import logging
import sys
from operator import methodcaller
from pathlib import Path

from boltons.ecoutils import get_profile
from boltons.iterutils import unique

# Canonical name of the CLI.
CLI_NAME = "mdedup"

__version__ = "6.1.1"


# Environment data.
env_data = get_profile(scrub=True)


# Initialize global logger.
logger = logging.getLogger(CLI_NAME)


# Ordered list of headers to use by default to compute the hash of a mail.
HASH_HEADERS = (
    "Date",
    "From",
    "To",
    # No Cc since mailman apparently sometimes trims list members from the Cc
    # header to avoid sending duplicates: https://mail.python.org/pipermail
    # /mailman-developers/2002-September/013233.html . But this means that
    # copies of mail reflected back from the list server will have a different
    # Cc to the copy saved by the MUA at send-time.
    # 'Cc',
    # No Bcc either since copies of the mail saved by the MUA at send-time
    # will have Bcc, but copies reflected back from the list server won't.
    # 'Bcc',
    # No Reply-To since a mail could be Cc'd to two lists with different
    # Reply-To munging options set.
    # 'Reply-To',
    "Subject",
    "MIME-Version",
    "Content-Type",
    "Content-Disposition",
    "User-Agent",
    "X-Priority",
    "Message-ID",
)


# Below this value, we consider not having enough data to compute a solid hash.
MINIMAL_HEADERS_COUNT = 4


# Since we're ignoring the Content-Length header for the reasons stated above,
# we limit the allowed difference between the sizes of the message payloads. If
# this is exceeded, a warning is issued and the messages are not considered
# duplicates, because this could point to message corruption somewhere, or a
# false positive. Note that the headers are not counted towards this threshold,
# because many headers can be added by mailing list software such as mailman,
# or even by the process of sending the mail through various MTAs - one copy
# could have been stored by the sender's MUA prior to sending, without any
# Received: headers, and another copy could be reflected back via a Cc-to-self
# mechanism or mailing list server. But this threshold has to be at least large
# enough to allow for footers added by mailing list servers.
DEFAULT_SIZE_THRESHOLD = 512  # bytes


# Similarly, we generated unified diffs of duplicates and ensure that the diff
# is not greater than a certain size.
DEFAULT_CONTENT_THRESHOLD = 768  # bytes


# Sources from which we compute a mail's canonical timestamp.
DATE_HEADER = "date-header"
CTIME = "ctime"
TIME_SOURCES = frozenset([DATE_HEADER, CTIME])


class TooFewHeaders(Exception):

    """ Not enough headers were found to produce a solid hash. """


class SizeDiffAboveThreshold(Exception):

    """ Difference in mail size is greater than threshold. """


class ContentDiffAboveThreshold(Exception):

    """ Difference in mail content is greater than threshold. """


class Config:

    """ Holds global configuration. """

    # Keep these defaults in sync with CLI option definitions.
    default_conf = {
        "dry_run": False,
        "input_format": False,
        "force_unlock": False,
        "hash_only": False,
        "hash_headers": HASH_HEADERS,
        "size_threshold": DEFAULT_SIZE_THRESHOLD,
        "content_threshold": DEFAULT_CONTENT_THRESHOLD,
        "show_diff": False,
        "strategy": None,
        "time_source": None,
        "regexp": None,
        "action": None,
        "export": None,
        "export_format": "mbox",
    }

    def __init__(self, **kwargs):
        """ Validates configuration parameter types and values. """
        # Load default values.
        self.conf = self.default_conf.copy()

        unrecognized_options = set(kwargs) - set(self.default_conf)
        if unrecognized_options:
            raise ValueError(f"Unrecognized {unrecognized_options} options.")

        # Replace defaults values with our config.
        self.conf.update(kwargs)

        # Check thresholds.
        assert self.size_threshold >= -1
        assert self.content_threshold >= -1

        # Headers are case-insensitive in Python implementation.
        normalized_headers = [h.lower() for h in self.hash_headers]
        # Remove duplicate entries.
        normalized_headers = unique(normalized_headers)
        # Mail headers are composed of ASCII characters between 33 and 126
        # (both inclusive) according the RFC-5322.
        for hid in normalized_headers:
            ascii_indexes = set(map(ord, hid))
            assert max(ascii_indexes) <= 126
            assert min(ascii_indexes) >= 33
        self.hash_headers = tuple(normalized_headers)

        # Export mail box will always be created from scratch and is not
        # expected to exists in the first place.
        if self.export:
            self.export = Path(self.export).resolve()
            if self.export.exists():
                raise FileExistsError(self.export)

    def __getattr__(self, attr_id):
        """ Expose configuration entries as properties. """
        if attr_id in self.conf:
            return self.conf[attr_id]



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from mailbox import Maildir
import logging
from pathlib import Path

import pytest

from .. import __version__, logger


def test_real_fs():
    """Check a simple test is not caught into the CLI runner fixture which is
    encapsulating all filesystem access into temporary directory structure."""
    assert str(Path(__file__)).startswith(str(Path.cwd()))


def test_temporary_fs(runner):
    """Check the CLI runner fixture properly encapsulated the filesystem in
    temporary directory."""
    assert not str(Path(__file__)).startswith(str(Path.cwd()))


def test_bare_call(invoke):
    result = invoke()
    assert result.exit_code == 0
    assert "Usage: " in result.output


def test_main_help(invoke):
    result = invoke("--help")
    assert result.exit_code == 0
    assert "Usage: " in result.output


def test_version(invoke):
    result = invoke("--version")
    assert result.exit_code == 0
    assert __version__ in result.output


def test_unknown_option(invoke):
    result = invoke("--blah")
    assert result.exit_code == 2
    assert "Error: no such option: --blah" in result.output


def test_unrecognized_verbosity(invoke):
    result = invoke("--verbosity", "random")
    assert result.exit_code == 2
    assert "Error: Invalid value for '--verbosity' / '-v'" in result.output


@pytest.mark.parametrize("level", ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"])
def test_verbosity(invoke, level):
    result = invoke("--verbosity", level)
    assert result.exit_code == 0
    assert "Usage: " in result.output

    assert logger.level == getattr(logging, level)
    if level == "DEBUG":
        assert "debug: " in result.output
    else:
        assert "debug: " not in result.output


def test_early_export_file_check(invoke, make_box, tmp_path):
    """Ensures the export file is tested for existence before any process is ran.

    See: https://github.com/kdeldycke/mail-deduplicate/issues/119"""
    box_path, _ = make_box(Maildir)

    result = invoke("--export=non_existing.file", box_path)
    assert result.exit_code == 0
    assert "0 mails found." in result.output
    assert "● Phase #0" in result.output
    assert "non_existing.file" not in result.output

    file = tmp_path.joinpath("existing.file")
    file.touch()
    result = invoke(f"--export={file!s}", box_path)
    assert result.exit_code == 1
    assert result.output == ""
    assert str(result.exception) == str(file)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from boltons.iterutils import unique
from boltons.dictutils import FrozenDict

from . import logger
from .colorize import choice_style
from .mailbox import create_box


# Actions performed on the mail selection.
COPY_SELECTED = "copy-selected"
COPY_DISCARDED = "copy-discarded"
MOVE_SELECTED = "move-selected"
MOVE_DISCARDED = "move-discarded"
DELETE_SELECTED = "delete-selected"
DELETE_DISCARDED = "delete-discarded"


def copy_selected(dedup):
    """Copy all mails selected to a brand new box."""
    box = create_box(dedup.conf.export, dedup.conf.export_format)

    for mail in dedup.selection:
        logger.debug(f"Copying {mail!r} to {dedup.conf.export}...")
        dedup.stats["mail_copied"] += 1
        if dedup.conf.dry_run:
            logger.warning("DRY RUN: Skip action.")
        else:
            box.add(mail)
            logger.info(f"{mail!r} copied.")

    logger.debug(f"Close {dedup.conf.export}")
    box.close()


def move_selected(dedup):
    """Move all mails selected to a brand new box."""
    box = create_box(dedup.conf.export, dedup.conf.export_format)

    for mail in dedup.selection:
        logger.debug(f"Move {mail!r} form {mail.source_path} to {dedup.conf.export}...")
        dedup.stats["mail_moved"] += 1
        if dedup.conf.dry_run:
            logger.warning("DRY RUN: Skip action.")
        else:
            box.add(mail)
            dedup.sources[mail.source_path].remove(mail.mail_id)
            logger.info(f"{mail!r} copied.")

    logger.debug(f"Close {dedup.conf.export}")
    box.close()


def delete_selected(dedup):
    """Remove all mails selected in-place, from their original boxes."""
    for mail in dedup.selection:
        logger.debug(f"Deleting {mail!r} in-place...")
        dedup.stats["mail_deleted"] += 1
        if dedup.conf.dry_run:
            logger.warning("DRY RUN: Skip action.")
        else:
            dedup.sources[mail.source_path].remove(mail.mail_id)
            logger.info(f"{mail!r} deleted.")


ACTIONS = FrozenDict(
    {
        COPY_SELECTED: copy_selected,
        COPY_DISCARDED: None,
        MOVE_SELECTED: move_selected,
        MOVE_DISCARDED: None,
        DELETE_SELECTED: delete_selected,
        DELETE_DISCARDED: None,
    }
)


def perform_action(dedup):
    """Performs the action on selected mail candidates."""
    logger.info(f"Perform {choice_style(dedup.conf.action)} action...")

    selection_count = len(dedup.selection)
    if selection_count == 0:
        logger.warning("No mail selected to perform action on.")
        return
    logger.info(f"{selection_count} mails selected for action.")

    # Check our indexing and selection methods are not flagging candidates
    # several times.
    assert len(unique(dedup.selection)) == len(dedup.selection)
    assert len(dedup.selection) == dedup.stats["mail_selected"]

    # Hunt down for action implementation.
    method = ACTIONS.get(dedup.conf.action)
    if not method:
        raise NotImplementedError(f"{dedup.conf.action} action not implemented yet.")

    method(dedup)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import email
import hashlib
import inspect
import mailbox
import os
import re
import time

import arrow
from boltons.cacheutils import cachedproperty
from tabulate import tabulate

from . import CTIME, MINIMAL_HEADERS_COUNT, TooFewHeaders, logger


class DedupMail:

    """Message with deduplication-specific properties and utilities.

    Adds all data-cleaning primitives and heuristics to Python's standard
    library messages from:
    https://github.com/python/cpython/blob/e799aa8b92c195735f379940acd9925961ad04ec/Lib/mailbox.py#L1489

    This class should not be used directly but composed with
    ``mailbox.Message`` sub-classes.
    """

    def __init__(self, message=None):
        """Initialize a pre-parsed ``Message`` instance the same way the default
        factory in Python's ``mailbox`` module does.
        """
        # Hunt down in our parent classes (but ourselve) the first one inheriting the
        # mailbox.Message class. That way we can get to the original factory.
        orig_message_klass = None
        for klass in inspect.getmro(self.__class__)[1:]:
            if issubclass(klass, mailbox.Message):
                orig_message_klass = klass
                break
        assert orig_message_klass

        # Call original object initialization from the right message class we
        # inherits from mailbox.Message.
        super(orig_message_klass, self).__init__(message)

        # Normalized path to the mailbox this message originates from.
        self.source_path = None

        # Mail ID used to uniquely refers to it in the context of its source.
        self.mail_id = None

        # Real filesystem location of the mail. Returns the individual mail's file
        # for folder-based box types (maildir & co.), but returns the whole box path
        # for file-based boxes (mbox & co.). Only used by regexp-based selection
        # strategies.
        self.path = None

        # Global config.
        self.conf = None

    def add_box_metadata(self, box, mail_id):
        """Post-instantiation utility to attach to mail some metadata derived from its
        parent box.

        Called right after the ``__init__()`` constructor.

        This allows the mail to carry its own information on its origin box and index.
        """
        self.source_path = box._path
        self.mail_id = mail_id

        # Extract file name and close it right away to reclaim memory.
        mail_file = box.get_file(mail_id)
        self.path = mail_file._file.name
        mail_file.close()

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.source_path}:{self.mail_id}>"

    @cachedproperty
    def uid(self):
        """Unique ID of the mail."""
        return self.source_path, self.mail_id

    @cachedproperty
    def timestamp(self):
        """Compute the normalized canonical timestamp of the mail.

        Sourced from the message's header by default. In the case of maildir,
        can be sourced from the email's file from the filesystem.
        """
        # XXX ctime does not refer to creation time on POSIX systems, but
        # rather the last time the inode data changed. Source:
        # https://userprimary.net/posts/2007/11/18
        # /ctime-in-unix-means-last-change-time-not-create-time/
        if self.conf.time_source == CTIME:
            return os.path.getctime(self.path)

        # Fetch from the date header.
        value = self.get("Date")
        try:
            value = email.utils.mktime_tz(email.utils.parsedate_tz(value))
        except ValueError:
            pass
        return value

        # XXX Also investigate what https://docs.python.org/2/library
        # /mailbox.html#mailbox.MaildirMessage.get_date does.

    @cachedproperty
    def size(self):
        """Returns canonical mail size.

        Size is computed as the length of the message body, i.e. the payload of
        the mail stripped of all its headers, not from the mail file
        persisting on the file-system.
        """
        return len("".join(self.body_lines))

        # TODO: Allow customization of the way the size is computed, by getting
        # the file size instead for example.
        # size = os.path.getsize(mail_file)

    @cachedproperty
    def body_lines(self):
        """ Return a normalized list of lines from message's body. """
        body = []
        if self.preamble is not None:
            body.extend(self.preamble.splitlines(keepends=True))

        for part in self.walk():
            if part.is_multipart():
                continue

            ctype = part.get_content_type()
            cte = part.get_params(header="Content-Transfer-Encoding")
            if (ctype is not None and not ctype.startswith("text")) or (
                cte is not None and cte[0][0].lower() == "8bit"
            ):
                part_body = part.get_payload(decode=False)
            else:
                charset = part.get_content_charset()
                if charset is None or len(charset) == 0:
                    charsets = ["ascii", "utf-8"]
                else:
                    charsets = [charset]

                part_body = part.get_payload(decode=True)
                for enc in charsets:
                    try:
                        part_body = part_body.decode(enc)
                        break
                    except UnicodeDecodeError:
                        continue
                    except LookupError:
                        continue
                else:
                    part_body = part.get_payload(decode=False)

            body.extend(part_body.splitlines(keepends=True))

        if self.epilogue is not None:
            body.extend(self.epilogue.splitlines(keepends=True))
        return body

    @cachedproperty
    def subject(self):
        """Normalized subject.

        Only used for debugging and human-friendly logging.
        """
        # Fetch subject from first message.
        subject = self.get("Subject", "")
        subject, _ = re.subn(r"\s+", " ", subject)
        return subject

    @cachedproperty
    def hash_key(self):
        """ Returns the canonical hash of a mail. """
        logger.debug(f"Serialized headers: {self.serialized_headers!r}")
        hash_value = hashlib.sha224(self.serialized_headers).hexdigest()
        logger.debug(f"Hash: {hash_value}")
        return hash_value

    @cachedproperty
    def canonical_headers(self):
        """Returns the full list of all canonical headers names and values in
        preparation for hashing."""
        canonical_headers = []

        for header_id in self.conf.hash_headers:

            # Skip absent header.
            if header_id not in self:
                continue

            # Fetch all occurrences of the header.
            canonical_values = []
            for header_value in self.get_all(header_id):
                normalized_value = self.normalize_header_value(header_id, header_value)
                if re.search(r"\S", normalized_value):
                    canonical_values.append(normalized_value)
            canonical_value = "\n".join(canonical_values)

            canonical_headers.append((header_id, canonical_value))

        # Cast to a tuple to prevent any modification.
        return tuple(canonical_headers)

    @cachedproperty
    def pretty_canonical_headers(self):
        """Renders into a table and in the same order, headers names and values
        used to produce mail's hash.

        Returns a string ready for printing to the user or for debugging.
        """
        table = [["Header ID", "Header value"]] + list(self.canonical_headers)
        return "\n" + tabulate(table, tablefmt="fancy_grid", headers="firstrow")

    @cachedproperty
    def serialized_headers(self):
        """ Serialize the canonical headers into a single string ready to be hashed. """
        # At this point we should have at an absolute minimum of headers.
        headers_count = len(self.canonical_headers)
        if headers_count < MINIMAL_HEADERS_COUNT:
            logger.warning(self.pretty_canonical_headers)
            raise TooFewHeaders(
                f"{headers_count} headers found out of {MINIMAL_HEADERS_COUNT}."
            )
        else:
            logger.debug(self.pretty_canonical_headers)

        return "\n".join(
            [f"{h_id}: {h_value}" for h_id, h_value in self.canonical_headers]
        ).encode("utf-8")

    @staticmethod
    def normalize_header_value(header_id, value):
        """Normalize and clean-up header value into its canonical form.

        Always returns a unicode string.
        """
        # Problematic when reading utf8 emails
        # this will ensure value is always string
        if isinstance(value, bytes):
            value = value.decode("utf-8", "replace")
        elif isinstance(value, email.header.Header):
            value = str(value)

        # Normalize white spaces.
        value = re.sub(r"\s+", " ", value).strip()

        # Trim Subject prefixes automatically added by mailing list software,
        # since the mail could have been cc'd to multiple lists, in which case
        # it will receive a different prefix for each, but this shouldn't be
        # treated as a real difference between duplicate mails.
        if header_id == "subject":
            subject = value
            while True:
                matching = re.match(
                    r"([Rr]e: )*(\[\w[\w_-]+\w\] )+(.+)", subject, re.DOTALL
                )
                if not matching:
                    break
                subject = matching.group(3)
                # show_progress("Trimmed Subject to %s" % subject)
            return subject

        if header_id == "content-type":
            # Apparently list servers actually munge Content-Type
            # e.g. by stripping the quotes from charset="us-ascii".
            # Section 5.1 of RFC2045 says that either form is valid
            # (and they are equivalent).
            #
            # Additionally, with multipart/mixed, boundary delimiters can
            # vary by recipient.  We need to allow for duplicates coming
            # from multiple recipients, since for example you could be
            # signed up to the same list twice with different addresses.
            # Or maybe someone bounces you a load of mail some of which is
            # from a mailing list you're both subscribed to - then it's
            # still useful to be able to eliminate duplicates.
            return re.sub(";.*", "", value)

        if header_id == "date":
            # Date timestamps can differ by seconds or hours for various
            # reasons, so let's only honour the date for now and normalize them
            # to UTC timezone.
            try:
                parsed = email.utils.parsedate_tz(value)
                if not parsed:
                    raise TypeError
                utc_timestamp = email.utils.mktime_tz(parsed)
                return arrow.get(utc_timestamp).format("YYYY-MM-DD")
            except (TypeError, ValueError):
                return value

        elif header_id in ["to", "message-id"]:
            # Sometimes email.parser strips the <> brackets from a To:
            # header which has a single address.  I have seen this happen
            # for only one mail in a duplicate pair.  I'm not sure why
            # (presumably the parser uses email.utils.unquote somewhere in
            # its code path which was only triggered by that mail and not
            # its sister mail), but to be safe, we should always strip the
            # <> brackets to avoid this difference preventing duplicate
            # detection.
            if re.match("^<[^<>,]+>$", value):
                return email.utils.unquote(value)

        return value



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import inspect
import mailbox
from functools import partial
from pathlib import Path

from boltons.dictutils import FrozenDict
from boltons.iterutils import flatten

from . import logger
from .colorize import choice_style
from .mail import DedupMail

""" Patch and tweak Python's standard librabry mailboxes constructors to set
sane defaults. Also forces out our own message factories to add deduplication
tools and utilities. """


def build_box_constructors():
    """Build our own mail constructors for each mailbox format.

    Gather all constructors defined by the standard Python library and extend them
    with our ``DedupMail`` class.
    """
    # Only keep subclasses of the ``mailbox.Mailbox`` interface, but the latter and
    # all others starting with an underscore.
    for _, klass in inspect.getmembers(mailbox, inspect.isclass):
        if (
            klass != mailbox.Mailbox
            and not klass.__name__.startswith("_")
            and issubclass(klass, mailbox.Mailbox)
        ):
            # Fetch the default factory for each mailbox type based on naming
            # conventions.
            message_klass = getattr(mailbox, f"{klass.__name__}Message")
            assert issubclass(message_klass, mailbox.Message)

            # Extend the default factory with DedupMail class.
            factory_klass = type(
                f"{klass.__name__}DedupMail",
                (DedupMail, message_klass, object),
                {
                    "__doc__": f"Extend the default message factory for {klass} with "
                    "our own ``DedupMail`` class to add deduplication utilities."
                },
            )

            # Set our own custom factory and safety options to default constructor.
            constructor = partial(klass, factory=factory_klass, create=False)

            # Generates our own box_type_id for use in CLI parameters.
            box_type_id = klass.__name__.lower()

            yield box_type_id, constructor


# Mapping between supported box type IDs and their constructors.
BOX_TYPES = FrozenDict(build_box_constructors())


# Categorize each box type into its structure type.
BOX_STRUCTURES = FrozenDict(
    {
        "file": {"mbox", "mmdf", "babyl"},
        "folder": {"maildir", "mh"},
    }
)
# Check we did not forgot any box type.
assert set(flatten(BOX_STRUCTURES.values())) == set(BOX_TYPES)


# List of required sub-folders defining a properly structured maildir.
MAILDIR_SUBDIRS = frozenset(("cur", "new", "tmp"))


def autodetect_box_type(path):
    """Auto-detect the format of the mailbox located at the provided path.

    Returns a box type as indexed in the ``box_types`` dictionnary above.

    If the path is a file, then it is considered as an ``mbox``. Else, if th
    provided path is a folder and feature the expecteed sub-directories, it is
    parsed as a ``maildir``.

    Future finer autodetection heuristics should be implemented here. Some ideas:
        * single mail from a maildir
        * plain text mail content
        * other mailbox formats supported in Python's std lib:
            * ``MH``
            * ``Babyl``
            * ``MMDF``
    """
    box_type = None

    # Validates folder is a maildir.
    if path.is_dir():
        for subdir in MAILDIR_SUBDIRS:
            if not path.joinpath(subdir).is_dir():
                raise ValueError(f"Missing sub-directory {subdir!r}")
        box_type = "maildir"

    # Validates folder is a mbox.
    elif path.is_file():
        box_type = "mbox"

    if not box_type:
        raise ValueError("Unrecognized mail source type.")

    logger.info(f"{choice_style(box_type)} detected.")
    return box_type


def open_box(path, box_type=False, force_unlock=False):
    """Open a mailbox.

    Returns a list of boxes, one per sub-folder. All are locked, ready for operations.

    If ``box_type`` is provided, forces the opening of the box in the specified format.
    Defaults to (crude) autodetection.
    """
    logger.info(f"\nOpening {choice_style(path)} ...")
    path = Path(path)
    if not box_type:
        box_type = autodetect_box_type(path)
    else:
        logger.warning(f"Forcing {box_type} format.")

    constructor = BOX_TYPES[box_type]
    # Do not allow the constructor to create a new mailbox if not found.
    box = constructor(path, create=False)

    return open_subfolders(box, force_unlock)


def lock_box(box, force_unlock):
    """Lock an opened box and allows for forced unlocking.

    Returns the locked box.
    """
    try:
        logger.debug("Locking box...")
        box.lock()
    except mailbox.ExternalClashError:
        logger.error("Box already locked!")
        # Remove the lock manually and re-lock.
        if force_unlock:
            logger.warning("Forcing removal of lock...")
            # Forces internal metadata.
            box._locked = True
            box.unlock()
            box.lock()
        # Re-raise error.
        else:
            raise
    logger.debug("Box opened.")
    return box


def open_subfolders(box, force_unlock):
    """Browse recursively the subfolder tree of a box.

    Returns a list of opened and locked boxes, each for one subfolder.
    """
    folder_list = [lock_box(box, force_unlock)]

    # Skip box types not supporting subfolders.
    if hasattr(box, "list_folders"):
        for folder_id in box.list_folders():
            logger.info(f"Opening subfolder {folder_id} ...")
            folder_list += open_subfolders(box.get_folder(folder_id), force_unlock)
    return folder_list


def create_box(path, box_type=False):
    """Creates a brand new box from scratch."""
    assert isinstance(path, Path)
    logger.info(
        f"Creating new {choice_style(box_type)} box at {choice_style(str(path))} ..."
    )

    if path.exists():
        raise FileExistsError(path)

    constructor = BOX_TYPES[box_type]
    # Allow the constructor to create a new mail box as we already double-checked
    # beforehand it does not exist.
    box = constructor(path, create=True)

    logger.debug("Locking box...")
    box.lock()
    return box



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Strategy definitions. """

import random
import re

from boltons.dictutils import FrozenDict

from . import logger


def select_older(duplicates):
    """Select all older duplicates.

    Discards the newests, i.e. the subset sharing the most recent timestamp.
    """
    logger.info(
        f"Select all mails strictly older than the {duplicates.newest_timestamp} "
        "timestamp..."
    )
    return {
        mail for mail in duplicates.pool if mail.timestamp < duplicates.newest_timestamp
    }


def select_oldest(duplicates):
    """Select all the oldest duplicates.

    Discards the newers, i.e. all mail of the duplicate set but those sharing
    the oldest timestamp.
    """
    logger.info(
        f"Select all mails sharing the oldest {duplicates.oldest_timestamp} "
        "timestamp..."
    )
    return {
        mail
        for mail in duplicates.pool
        if mail.timestamp == duplicates.oldest_timestamp
    }


def select_newer(duplicates):
    """Select all newer duplicates.

    Discards the oldest, i.e. the subset sharing the most ancient timestamp.
    """
    logger.info(
        f"Select all mails strictly newer than the {duplicates.oldest_timestamp} "
        "timestamp..."
    )
    return {
        mail for mail in duplicates.pool if mail.timestamp > duplicates.oldest_timestamp
    }


def select_newest(duplicates):
    """Select all the newest duplicates.

    Discards the olders, i.e. all mail of the duplicate set but those sharing the
    newest timestamp.
    """
    logger.info(
        f"Select all mails sharing the newest {duplicates.newest_timestamp} "
        "timestamp..."
    )
    return {
        mail
        for mail in duplicates.pool
        if mail.timestamp == duplicates.newest_timestamp
    }


def select_smaller(duplicates):
    """Select all smaller duplicates.

    Discards the biggests, i.e. the subset sharing the biggest size.
    """
    logger.info(
        f"Select all mails strictly smaller than {duplicates.biggest_size} bytes..."
    )
    return {mail for mail in duplicates.pool if mail.size < duplicates.biggest_size}


def select_smallest(duplicates):
    """Select all the smallest duplicates.

    Discards the biggers. i.e. all mail of the duplicate set but those sharing the
    smallest size.
    """
    logger.info(
        f"Select all mails sharing the smallest size of {duplicates.smallest_size} "
        "bytes..."
    )
    return {mail for mail in duplicates.pool if mail.size == duplicates.smallest_size}


def select_bigger(duplicates):
    """Select all bigger duplicates.

    Discards the smallests, i.e. the subset sharing the smallest size.
    """
    logger.info(
        f"Select all mails strictly bigger than {duplicates.smallest_size} bytes..."
    )
    return {mail for mail in duplicates.pool if mail.size > duplicates.smallest_size}


def select_biggest(duplicates):
    """Select all the biggest duplicates.

    Discards the smallers, i.e. all mail of the duplicate set but those sharing the
    biggest size.
    """
    logger.info(
        f"Select all mails sharing the biggest size of {duplicates.biggest_size} "
        "bytes..."
    )
    return {mail for mail in duplicates.pool if mail.size == duplicates.biggest_size}


def select_matching_path(duplicates):
    """Select all duplicates whose file path match the regular expression provided
    via the --regexp parameter.
    """
    logger.info(
        "Select all mails with file path matching the "
        f"{duplicates.conf.regexp.pattern} regexp..."
    )
    return {
        mail for mail in duplicates.pool if re.search(duplicates.conf.regexp, mail.path)
    }


def select_non_matching_path(duplicates):
    """Select all duplicates whose file path doesn't match the regular expression
    provided via the --regexp parameter.
    """
    logger.info(
        "Select all mails with file path not matching the "
        f"{duplicates.conf.regexp.pattern} regexp..."
    )
    return {
        mail
        for mail in duplicates.pool
        if not re.search(duplicates.conf.regexp, mail.path)
    }


def select_one(duplicates):
    """Randomly select one duplicate, and discards all others."""
    return {random.choice(tuple(duplicates.pool))}


def select_all_but_one(duplicates):
    """Randomly discard one duplicate, and select all others."""
    return set(random.sample(duplicates.pool, k=len(duplicates.pool) - 1))


# Use symbols to define selection strategies.
DISCARD_OLDER = "discard-older"
DISCARD_OLDEST = "discard-oldest"
DISCARD_NEWER = "discard-newer"
DISCARD_NEWEST = "discard-newest"
SELECT_OLDER = "select-older"
SELECT_OLDEST = "select-oldest"
SELECT_NEWER = "select-newer"
SELECT_NEWEST = "select-newest"

DISCARD_SMALLER = "discard-smaller"
DISCARD_SMALLEST = "discard-smallest"
DISCARD_BIGGER = "discard-bigger"
DISCARD_BIGGEST = "discard-biggest"
SELECT_SMALLER = "select-smaller"
SELECT_SMALLEST = "select-smallest"
SELECT_BIGGER = "select-bigger"
SELECT_BIGGEST = "select-biggest"

DISCARD_MATCHING_PATH = "discard-matching-path"
DISCARD_NON_MATCHING_PATH = "discard-non-matching-path"
SELECT_MATCHING_PATH = "select-matching-path"
SELECT_NON_MATCHING_PATH = "select-non-matching-path"

DISCARD_ONE = "discard-one"
DISCARD_ALL_BUT_ONE = "discard-all-but-one"
SELECT_ONE = "select-one"
SELECT_ALL_BUT_ONE = "select-all-but-one"


# Groups strategy aliases and their definitions. Aliases are great useability
# features as it helps users to better reason about the selection operators
# dependening on their mental models.
STRATEGY_ALIASES = frozenset(
    [
        (SELECT_NEWEST, DISCARD_OLDER),
        (SELECT_NEWER, DISCARD_OLDEST),
        (SELECT_OLDEST, DISCARD_NEWER),
        (SELECT_OLDER, DISCARD_NEWEST),
        (SELECT_BIGGEST, DISCARD_SMALLER),
        (SELECT_BIGGER, DISCARD_SMALLEST),
        (SELECT_SMALLEST, DISCARD_BIGGER),
        (SELECT_SMALLER, DISCARD_BIGGEST),
        (SELECT_NON_MATCHING_PATH, DISCARD_MATCHING_PATH),
        (SELECT_MATCHING_PATH, DISCARD_NON_MATCHING_PATH),
        (SELECT_ALL_BUT_ONE, DISCARD_ONE),
        (SELECT_ONE, DISCARD_ALL_BUT_ONE),
    ]
)


def get_method_id(strat_id):
    """ Transform strategy ID to its method ID. """
    return strat_id.replace("-", "_")


def build_method_mapping():
    """Precompute the mapping of all strategy IDs to their prefered method name,
    including aliases as fallbacks.
    """
    methods = dict()
    for strategies in STRATEGY_ALIASES:
        fallback_method = None
        for strat_id in strategies:
            mid = get_method_id(strat_id)
            method = globals().get(mid)
            if method:
                fallback_method = method
            if not fallback_method:
                raise NotImplementedError(f"Can't find {mid}() method.")
            methods[strat_id] = fallback_method
    return methods


STRATEGY_METHODS = FrozenDict(build_method_mapping())


def apply_strategy(strat_id, duplicates):
    """Perform the selection strategy on the provided duplicate set.

    Returns a set of selected mails objects.
    """
    if strat_id not in STRATEGY_METHODS:
        raise ValueError(f"Unknown {strat_id} strategy.")
    method = STRATEGY_METHODS[strat_id]
    logger.debug(f"Apply {method!r}...")
    return set(method(duplicates))



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Utilities to manage colors. """

import re
from functools import partial

import click
from click_log.core import ColorFormatter

from . import CLI_NAME, logger


# Help screen and log message colors.
# Available options: https://click.palletsprojects.com/en/master/api/#click.style
colors = {
    "cli": dict(fg="bright_white"),
    "title": dict(fg="bright_green", bold=True),
    "subtitle": dict(fg="green"),
    "option": dict(fg="cyan"),
    "choice": dict(fg="magenta"),
    "metavar": dict(fg="bright_black"),
}

# Update with default log message colors.
assert set(colors).isdisjoint(ColorFormatter.colors)
colors.update(ColorFormatter.colors)


# Reverse color mapping to detect overlapping colors.
color_index = {}
for category, color_params in colors.items():
    color_index.setdefault(tuple(sorted(color_params.items())), set()).add(category)
for color_params, categories in color_index.items():
    if len(categories) > 1:
        logger.debug(f"{categories!r} shares the same {color_params!r} colors.")


# Create handy function shortcuts for each class of colorization.
for color_category, color_params in colors.items():
    globals()[f"{color_category}_style"] = partial(click.style, **color_params)


def collect_keywords(ctx):
    """ Parse click context to collect option names and choice keywords. """
    options = set()
    choices = set()
    metavars = set()

    # Add user defined help options.
    options.update(ctx.help_option_names)

    # Collect all option names and choice keywords.
    for param in ctx.command.params:
        options.update(param.opts)
        if isinstance(param.type, click.Choice):
            choices.update(param.type.choices)
        if param.metavar:
            metavars.add(param.metavar)

    return options, choices, metavars


def colorized_help(help_txt, keywords):
    """Get default help screen and colorize section titles, options and choice
    keywords."""
    options, choices, metavars = keywords

    def colorize(match, **kwargs):
        """Re-create the matching string by concatenating all groups, but only
        colorize named groups.
        """
        txt = ""
        for group in match.groups():
            if group in match.groupdict().values():
                txt += click.style(group, **kwargs)
            else:
                txt += group
        return txt

    # Highligh numbers.
    help_txt = re.sub(
        r"(\s)(?P<colorize>-?\d+)", partial(colorize, **colors["choice"]), help_txt
    )

    # Highlight CLI.
    help_txt = re.sub(
        fr"(\s)(?P<colorize>{CLI_NAME})", partial(colorize, **colors["cli"]), help_txt
    )

    # Highligh sections.
    help_txt = re.sub(
        r"^(?P<colorize>\S[\S+ ]+)(:)",
        partial(colorize, **colors["title"]),
        help_txt,
        flags=re.MULTILINE,
    )

    # Highlight keywords.
    for matching_keywords, color in [
        (sorted(options), colors["option"]),
        (sorted(choices, reverse=True), colors["choice"]),
        (sorted(metavars, reverse=True), colors["metavar"]),
    ]:
        for keyword in matching_keywords:
            # Accounts for text wrapping after a dash.
            keyword = keyword.replace("-", "-\\s*")
            help_txt = re.sub(
                fr"([\s\[\|\(])(?P<colorize>{keyword})",
                partial(colorize, **color),
                help_txt,
            )

    return help_txt



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from collections import Counter, OrderedDict
from difflib import unified_diff
from itertools import combinations
from operator import attrgetter
from pathlib import Path
import textwrap

import click
from boltons.cacheutils import cachedproperty
from tabulate import tabulate

from . import ContentDiffAboveThreshold, SizeDiffAboveThreshold, TooFewHeaders, logger
from .colorize import choice_style, subtitle_style
from .mailbox import open_box
from .strategy import apply_strategy


# Reference all tracked statistics and their definition.
STATS_DEF = OrderedDict(
    [
        ("mail_found", "Total number of mails encountered from all mail sources."),
        (
            "mail_rejected",
            "Number of mails rejected individually because they were unparseable or "
            "did not have enough metadata to compute hashes.",
        ),
        (
            "mail_retained",
            "Number of valid mails parsed and retained for deduplication.",
        ),
        ("mail_hashes", "Number of unique hashes."),
        (
            "mail_unique",
            "Number of unique mails (which where automatically added to selection).",
        ),
        (
            "mail_duplicates",
            "Number of duplicate mails (sum of mails in all duplicate sets with at "
            "least 2 mails).",
        ),
        (
            "mail_skipped",
            "Number of mails ignored in the selection phase because the whole set "
            "they belong to was skipped.",
        ),
        ("mail_discarded", "Number of mails discarded from the final selection."),
        (
            "mail_selected",
            "Number of mails kept in the final selection on which the "
            "action will be performed.",
        ),
        (
            "mail_copied",
            "Number of mails copied from their original mailbox to another.",
        ),
        ("mail_moved", "Number of mails moved from their original mailbox to another."),
        ("mail_deleted", "Number of mails deleted from their mailbox in-place."),
        ("set_total", "Total number of duplicate sets."),
        (
            "set_single",
            "Total number of sets containing only a single mail with no applicable "
            "strategy. They were automatically kept in the final selection.",
        ),
        (
            "set_skipped_encoding",
            "Number of sets skipped from the selection process because they had "
            "encoding issues.",
        ),
        (
            "set_skipped_size",
            "Number of sets skipped from the selection process because they were "
            "too dissimilar in size.",
        ),
        (
            "set_skipped_content",
            "Number of sets skipped from the selection process because they were "
            "too dissimilar in content.",
        ),
        (
            "set_skipped_strategy",
            "Number of sets skipped from the selection process because the strategy "
            "could not be applied.",
        ),
        (
            "set_deduplicated",
            "Number of valid sets on which the selection strategy was successfully "
            "applied.",
        ),
    ]
)


class DuplicateSet:

    """A duplicate set of mails sharing the same hash.

    Implements all the safety checks required before we can apply any selection
    strategy.
    """

    def __init__(self, hash_key, mail_set, conf):
        """Load-up the duplicate set of mail and freeze pool.

        Once loaded-up, the pool of parsed mails is considered frozen for the
        rest of the duplicate set's life. This allows aggressive caching of
        lazy instance attributes depending on the pool content.
        """
        self.hash_key = hash_key

        # Global config.
        self.conf = conf

        # Pool referencing all duplicated mails and their attributes.
        self.pool = frozenset(mail_set)
        # There is no point creating a duplicate set with a single mail.
        assert self.size > 1

        # Set metrics.
        self.stats = Counter()
        self.stats["mail_duplicates"] += self.size

        logger.debug(f"{self!r} created.")

    def __repr__(self):
        """ Print internal raw states for debugging. """
        return f"<{self.__class__.__name__} hash={self.hash_key} size={self.size}>"

    @cachedproperty
    def size(self):
        """ Return the size of the duplicate set. """
        return len(self.pool)

    @cachedproperty
    def newest_timestamp(self):
        return max(map(attrgetter("timestamp"), self.pool))

    @cachedproperty
    def oldest_timestamp(self):
        return min(map(attrgetter("timestamp"), self.pool))

    @cachedproperty
    def biggest_size(self):
        return max(map(attrgetter("size"), self.pool))

    @cachedproperty
    def smallest_size(self):
        return min(map(attrgetter("size"), self.pool))

    def check_differences(self):
        """Ensures all mail differs in the limits imposed by size and content
        thresholds.

        Compare all mails of the duplicate set with each other, both in size
        and content. Raise an error if we're not within the limits imposed by
        the threshold settings.
        """
        logger.info("Check mail differences are below the thresholds.")
        if self.conf.size_threshold < 0:
            logger.info("Skip checking for size differences.")
        if self.conf.content_threshold < 0:
            logger.info("Skip checking for content differences.")
        if self.conf.size_threshold < 0 and self.conf.content_threshold < 0:
            return

        # Compute differences of mail against one another.
        for mail_a, mail_b in combinations(self.pool, 2):

            # Compare mails on size.
            if self.conf.size_threshold > -1:
                size_difference = abs(mail_a.size - mail_b.size)
                logger.debug(
                    f"{mail_a!r} and {mail_b!r} differs by {size_difference} bytes "
                    "in size."
                )
                if size_difference > self.conf.size_threshold:
                    raise SizeDiffAboveThreshold

            # Compare mails on content.
            if self.conf.content_threshold > -1:
                content_difference = self.diff(mail_a, mail_b)
                logger.debug(
                    f"{mail_a!r} and {mail_b!r} differs by {content_difference} bytes "
                    "in content."
                )
                if content_difference > self.conf.content_threshold:
                    if self.conf.show_diff:
                        logger.info(self.pretty_diff(mail_a, mail_b))
                    raise ContentDiffAboveThreshold

    def diff(self, mail_a, mail_b):
        # TODO: rewrite the diff algorithm to not rely on naive unified diff
        # result parsing.
        """Return difference in bytes between two mails' normalized body."""
        return len(
            "".join(
                unified_diff(
                    mail_a.body_lines,
                    mail_b.body_lines,
                    # Ignore difference in filename lengths and timestamps.
                    fromfile="a",
                    tofile="b",
                    fromfiledate="",
                    tofiledate="",
                    n=0,
                    lineterm="\n",
                )
            )
        )

    def pretty_diff(self, mail_a, mail_b):
        """Returns a verbose unified diff between two mails' normalized body."""
        return "".join(
            unified_diff(
                mail_a.body_lines,
                mail_b.body_lines,
                fromfile=f"Normalized body of {mail_a!r}",
                tofile=f"Normalized body of {mail_b!r}",
                fromfiledate=f"{mail_a.timestamp:0.2f}",
                tofiledate=f"{mail_b.timestamp:0.2f}",
                n=0,
                lineterm="\n",
            )
        )

    def select_candidates(self):
        """Returns the list of duplicates selected for action.

        Run preliminary checks and return the candidates fitting the configured
        strategy and safety checks.
        """
        # Fine-grained checks on mail differences.
        try:
            self.check_differences()
        except UnicodeDecodeError as expt:
            logger.warning("Skip set: unparseable mails due to bad encoding.")
            logger.debug(f"{expt}")
            self.stats["mail_skipped"] += self.size
            self.stats["set_skipped_encoding"] += 1
            return
        except SizeDiffAboveThreshold:
            logger.warning("Skip set: mails are too dissimilar in size.")
            self.stats["mail_skipped"] += self.size
            self.stats["set_skipped_size"] += 1
            return
        except ContentDiffAboveThreshold:
            logger.warning("Skip set: mails are too dissimilar in content.")
            self.stats["mail_skipped"] += self.size
            self.stats["set_skipped_content"] += 1
            return

        if not self.conf.strategy:
            logger.warning("Skip set: no strategy to apply.")
            self.stats["mail_skipped"] += self.size
            self.stats["set_skipped_strategy"] += 1
            return

        # Fetch the subset of selected mails from the set by applying strategy.
        selected = apply_strategy(self.conf.strategy, self)
        candidate_count = len(selected)

        # Duplicate sets matching as a whole are skipped altogether.
        if candidate_count == self.size:
            logger.warning(
                f"Skip set: all {candidate_count} mails within were selected. "
                "The strategy criterion was not able to discard some."
            )
            self.stats["mail_skipped"] += self.size
            self.stats["set_skipped_strategy"] += 1
            return

        logger.info(f"{candidate_count} mail candidates selected for action.")
        self.stats["mail_selected"] += candidate_count
        self.stats["mail_discarded"] += self.size - candidate_count
        self.stats["set_deduplicated"] += 1
        return selected


class Deduplicate:

    """Load-up messages, search for duplicates, apply selection strategy and perform
    the action.

    Similar messages sharing the same hash are grouped together in a ``DuplicateSet``.
    """

    def __init__(self, conf):
        # Index of mail sources by their full, normalized path. So we can refer
        # to them in Mail instances. Also have the nice side effect of natural
        # deduplication of sources themselves.
        self.sources = {}

        # All mails grouped by hashes.
        self.mails = {}

        # Mails selected after application of selection strategy.
        self.selection = set()

        # Global config.
        self.conf = conf

        # Deduplication statistics.
        self.stats = Counter(dict.fromkeys(STATS_DEF, 0))

    def add_source(self, source_path):
        """Registers a source of mails, validates and opens it. """
        # Make the path absolute, resolving any symlinks. Do not allow duplicates in
        # our sources, as we use the path as a unique key to tie back a mail from its
        # source when performing the action later.
        source_path = str(Path(source_path).resolve(strict=True))
        if source_path in self.sources:
            raise ValueError(f"{source_path} already added.")

        # Open and register the mail source. Subfolders will be registered as their
        # own box.
        boxes = open_box(source_path, self.conf.input_format, self.conf.force_unlock)
        for box in boxes:
            self.sources[box._path] = box

            # Track global mail count.
            mail_found = len(box)
            logger.info(f"{mail_found} mails found.")
            self.stats["mail_found"] += mail_found

    def hash_all(self):
        """Browse all mails from all registered sources, compute hashes and group mails
        by hash.

        Displays a progress bar as the operation might be slow.
        """
        logger.info(
            f"Use [{', '.join(map(choice_style, self.conf.hash_headers))}] headers to "
            "compute hashes."
        )

        with click.progressbar(
            length=self.stats["mail_found"],
            label="Hashed mails",
            show_pos=True,
        ) as progress:

            for box in self.sources.values():
                for mail_id, mail in box.iteritems():

                    mail.add_box_metadata(box, mail_id)

                    mail.conf = self.conf

                    try:
                        mail_hash = mail.hash_key
                    except TooFewHeaders as expt:
                        logger.warning(f"Rejecting {mail!r}: {expt.args[0]}")
                        self.stats["mail_rejected"] += 1
                    else:
                        # Use a set to deduplicate entries pointing to the same file.
                        self.mails.setdefault(mail_hash, set()).add(mail)
                        self.stats["mail_retained"] += 1

                    progress.update(1)

        self.stats["mail_hashes"] += len(self.mails)

    def select_all(self):
        """Gather the final selection of mails from each duplicate set.

        We apply the selection strategy one duplicate set at a time to keep memory
        footprint low and make the log easier to read.
        """
        if self.conf.strategy:
            logger.info(
                f"{choice_style(self.conf.strategy)} strategy will be applied on each "
                "duplicate set to select candidates."
            )
        else:
            logger.warning("No strategy configured, skip selection.")

        self.stats["set_total"] = len(self.mails)

        for hash_key, mail_set in self.mails.items():

            # Alter log level depending on set length.
            mail_count = len(mail_set)
            log_level = logger.debug if mail_count == 1 else logger.info
            log_level(subtitle_style(f"◼ {mail_count} mails sharing hash {hash_key}"))

            # Unique mails are always selected. No need to mobilize the whole
            # DuplicateSet machinery.
            if mail_count == 1:
                logger.debug("Add unique message to selection.")
                self.stats["mail_unique"] += 1
                self.stats["mail_selected"] += 1
                self.stats["set_single"] += 1
                candidates = mail_set

            # We need to resort to a selection strategy to discriminate mails
            # within the set.
            else:
                duplicates = DuplicateSet(hash_key, mail_set, self.conf)
                candidates = duplicates.select_candidates()
                # Merge duplicate set's stats to global stats.
                self.stats += duplicates.stats

            if candidates:
                self.selection.update(candidates)

    def close_all(self):
        """ Close all open boxes. """
        for source_path, box in self.sources.items():
            logger.debug(f"Close {source_path}")
            box.close()

    def report(self):
        """ Returns a text report of user-friendly statistics and metrics. """
        output = ""
        for prefix, title in (("mail_", "Mails"), ("set_", "Duplicate sets")):
            table = [[title, "Metric", "Description"]]
            for stat_id, desc in STATS_DEF.items():
                if stat_id.startswith(prefix):
                    table.append(
                        [
                            stat_id[len(prefix) :].replace("_", " - ").title(),
                            self.stats[stat_id],
                            "\n".join(textwrap.wrap(desc, 60)),
                        ]
                    )
            output += tabulate(table, tablefmt="fancy_grid", headers="firstrow")
            output += "\n"
        return output

    def check_stats(self):
        """Perform some high-level consistency checks on metrics.

        Helps users reports tricky edge-cases.
        """
        # Box opening stats.
        assert self.stats["mail_found"] >= self.stats["mail_rejected"]
        assert self.stats["mail_found"] >= self.stats["mail_retained"]
        assert self.stats["mail_found"] == (
            self.stats["mail_rejected"] + self.stats["mail_retained"]
        )
        # Mail grouping by hash.
        assert self.stats["mail_retained"] >= self.stats["mail_unique"]
        assert self.stats["mail_retained"] >= self.stats["mail_duplicates"]
        assert self.stats["mail_retained"] == (
            self.stats["mail_unique"] + self.stats["mail_duplicates"]
        )
        # Mail selection stats.
        assert self.stats["mail_retained"] >= self.stats["mail_skipped"]
        assert self.stats["mail_retained"] >= self.stats["mail_discarded"]
        assert self.stats["mail_retained"] >= self.stats["mail_selected"]
        assert self.stats["mail_retained"] == (
            self.stats["mail_skipped"]
            + self.stats["mail_discarded"]
            + self.stats["mail_selected"]
        )
        # Action stats.
        assert self.stats["mail_selected"] >= self.stats["mail_copied"]
        assert self.stats["mail_selected"] >= self.stats["mail_moved"]
        assert self.stats["mail_selected"] >= self.stats["mail_deleted"]
        assert self.stats["mail_selected"] in (
            self.stats["mail_copied"],
            self.stats["mail_moved"],
            self.stats["mail_deleted"],
        )
        # Sets accounting.
        assert self.stats["set_total"] == self.stats["mail_hashes"]
        assert self.stats["set_single"] == self.stats["mail_unique"]
        assert self.stats["set_total"] == (
            self.stats["set_single"]
            + self.stats["set_skipped_encoding"]
            + self.stats["set_skipped_size"]
            + self.stats["set_skipped_content"]
            + self.stats["set_skipped_strategy"]
            + self.stats["set_deduplicated"]
        )



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import pytest


@pytest.mark.parametrize("source", ["./dummy_maildir/", "./__init__.py"])
def test_nonexistent_path(invoke, source):
    result = invoke(source)
    assert result.exit_code == 2
    assert f"Path '{source}' does not exist" in result.output


def test_invalid_maildir_structure(invoke):
    result = invoke("--action=delete-discarded", ".")
    assert result.exit_code == 1
    assert "Phase #0" in result.output
    assert "Opening " in result.output
    assert "Missing sub-directory" in str(result.exc_info[1])



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import logging
import re

import click
import click_log
from click_help_colors import version_option

from . import (
    CLI_NAME,
    DATE_HEADER,
    DEFAULT_CONTENT_THRESHOLD,
    DEFAULT_SIZE_THRESHOLD,
    HASH_HEADERS,
    TIME_SOURCES,
    Config,
    __version__,
    env_data,
    logger,
)
from .colorize import (
    collect_keywords,
    colorized_help,
    title_style,
    choice_style,
    colors,
)
from .deduplicate import Deduplicate
from .mailbox import BOX_TYPES, BOX_STRUCTURES
from .action import (
    ACTIONS,
    perform_action,
    COPY_SELECTED,
    COPY_DISCARDED,
    MOVE_SELECTED,
    MOVE_DISCARDED,
)
from .strategy import (
    DISCARD_MATCHING_PATH,
    DISCARD_NON_MATCHING_PATH,
    SELECT_MATCHING_PATH,
    SELECT_NON_MATCHING_PATH,
    STRATEGY_METHODS,
)

click_log.basic_config(logger)


def validate_regexp(ctx, param, value):
    """ Validate and compile regular expression. """
    if value:
        try:
            value = re.compile(value)
        except ValueError:
            raise click.BadParameter("invalid regular expression.")
    return value


@click.command(
    short_help="Deduplicate mail boxes.",
    no_args_is_help=False,
)
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
    default=False,
    help="Do not perform any action but act as if it was, and report which action "
    "would have been performed otherwise.",
)
@click.option(
    "-i",
    "--input-format",
    type=click.Choice(sorted(BOX_TYPES), case_sensitive=False),
    help="Force all provided mail sources to be parsed in the specified format. If "
    "not set, auto-detect the format of sources independently. Auto-detection only "
    "supports maildir and mbox format. Use this option to open up other box "
    "format, or bypass unreliable detection.",
)
@click.option(
    "-u",
    "--force-unlock",
    is_flag=True,
    default=False,
    help="Remove the lock on mail source opening if one is found.",
)
@click.option(
    "-H",
    "--hash-only",
    is_flag=True,
    default=False,
    help="Compute and display the internal hashes used to identify duplicates. Do not "
    "performs any selection or action.",
)
@click.option(
    "-h",
    "--hash-header",
    multiple=True,
    type=str,
    metavar="Header-ID",
    default=HASH_HEADERS,
    help="Headers to use to compute each mail's hash. Must be repeated multiple times "
    "to set an ordered list of headers. Header IDs are case-insensitive. Repeating "
    "entries are ignored. Defaults to: {}.".format(
        " ".join([f"-h {choice_style(h)}" for h in HASH_HEADERS])
    ),
)
@click.option(
    "-S",
    "--size-threshold",
    type=int,
    metavar="BYTES",
    default=DEFAULT_SIZE_THRESHOLD,
    help="Maximum difference allowed in size between mails sharing the same hash. "
    "The whole subset of duplicates will be skipped if at least one pair of mail "
    "exceeds the threshold. Set to 0 to enforce strictness and apply selection strategy "
    "on the subset only if all mails are exactly the same. Set to -1 to allow any "
    "difference and apply the strategy whatever the differences. Defaults to "
    f"{DEFAULT_SIZE_THRESHOLD} bytes.",
)
@click.option(
    "-C",
    "--content-threshold",
    type=int,
    metavar="BYTES",
    default=DEFAULT_CONTENT_THRESHOLD,
    help="Maximum difference allowed in content between mails sharing the same hash. "
    "The whole subset of duplicates will be skipped if at least one pair of mail "
    "exceeds the threshold. Set to 0 to enforce strictness and apply selection strategy "
    "on the subset only if all mails are exactly the same. Set to -1 to allow any "
    "difference and apply the strategy whatever the differences. Defaults to "
    f"{DEFAULT_CONTENT_THRESHOLD} bytes.",
)
@click.option(
    "-d",
    "--show-diff",
    is_flag=True,
    default=False,
    help="Show the unified diff of duplicates not within thresholds.",
)
@click.option(
    "-s",
    "--strategy",
    type=click.Choice(sorted(STRATEGY_METHODS), case_sensitive=False),
    help="Selection strategy to apply within a subset of duplicates. If not set, "
    "duplicates will be grouped and counted but all be skipped, selection will be "
    "empty, and no action will be performed. Description of each strategy is "
    "available at the bottom.",
)
@click.option(
    "-t",
    "--time-source",
    default=DATE_HEADER,
    type=click.Choice(sorted(TIME_SOURCES), case_sensitive=False),
    help="Source of a mail's time reference used in time-sensitive strategies. "
    f"Defaults to {DATE_HEADER}.",
)
@click.option(
    "-r",
    "--regexp",
    callback=validate_regexp,
    metavar="REGEXP",
    help="Regular expression on a mail's file path. Applies to real, individual "
    "mail location for folder-based boxed "
    f"({', '.join(sorted(BOX_STRUCTURES['folder']))}). But for file-based boxes "
    f"({', '.join(sorted(BOX_STRUCTURES['file']))}), applies to the whole box's path, "
    "as all mails are packed into one single file. Required in "
    f"{DISCARD_MATCHING_PATH}, {DISCARD_NON_MATCHING_PATH}, {SELECT_MATCHING_PATH} and "
    f"{SELECT_NON_MATCHING_PATH} strategies.",
)
@click.option(
    "-a",
    "--action",
    default=COPY_SELECTED,
    type=click.Choice(sorted(ACTIONS), case_sensitive=False),
    help=f"Action performed on the selected mails. Defaults to {COPY_SELECTED} as it "
    "is the safest: it only reads the mail sources and create a brand new mail box "
    "with the selection results.",
)
@click.option(
    "-E",
    "--export",
    metavar="MAIL_BOX_PATH",
    type=click.Path(resolve_path=True),
    help="Location of the destination mail box to where to copy or move deduplicated "
    f"mails. Required in {COPY_SELECTED}, {COPY_DISCARDED}, "
    f"{MOVE_SELECTED} and {MOVE_DISCARDED} actions.",
)
@click.option(
    "-e",
    "--export-format",
    default="mbox",
    type=click.Choice(sorted(BOX_TYPES), case_sensitive=False),
    help="Format of the mail box to which deduplication mails will be exported to. "
    f"Defaults to mbox. Only affects {COPY_SELECTED}, {COPY_DISCARDED}, "
    f"{MOVE_SELECTED} and {MOVE_DISCARDED} actions.",
)
@click.argument(
    "mail_sources",
    nargs=-1,
    metavar="MAIL_SOURCE_1 MAIL_SOURCE_2 ...",
    type=click.Path(exists=True, resolve_path=True),
)
@click_log.simple_verbosity_option(
    logger,
    default="INFO",
    metavar="LEVEL",
    type=click.Choice(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"], case_sensitive=False
    ),
    help="Either CRITICAL, ERROR, WARNING, INFO or DEBUG. Defaults to INFO.",
)
@version_option(
    version=__version__,
    prog_name=CLI_NAME,
    version_color="green",
    prog_name_color=colors["cli"]["fg"],
    message=f"%(prog)s %(version)s\n{env_data}",
    message_color="bright_black",
)
@click.pass_context
def mdedup(
    ctx,
    dry_run,
    input_format,
    force_unlock,
    hash_only,
    hash_header,
    size_threshold,
    content_threshold,
    show_diff,
    strategy,
    time_source,
    regexp,
    action,
    export,
    export_format,
    mail_sources,
):
    """Deduplicate mails from a set of mail boxes.

    \b
    Process:
    * Phase #1: run a first pass to compute from their headers the canonical hash of
                each encountered mail.
    * Phase #2: a second pass to apply the selection strategy on each subset of
                duplicate mails sharing the same hash.
    * Phase #3: perform an action on all selected mails.

    Action on the selected mails in phase #3 is only performed if no major differences
    between mails are uncovered during a fine-grained check differences in the second
    phase. Limits can be set via the --size-threshold and --content-threshold
    options, and are used as safety checks to prevent slightly different mails
    to be seen as similar through the lens of normalization.
    """
    level = logger.level
    level_name = logging._levelToName.get(level, level)
    logger.debug(f"Verbosity set to {level_name}.")

    # Print help screen and exit if no mail source provided.
    if not mail_sources:
        # Extract keywords
        keywords = collect_keywords(ctx)

        # Apply dynamic style to help screen.
        click.echo(colorized_help(ctx.get_help(), keywords))

        # Produce the strategy reference table, with grouped aliases.
        method_to_ids = {}
        for strat_id, method in sorted(STRATEGY_METHODS.items(), reverse=True):
            method_to_ids.setdefault(method, []).append(strat_id)
        strat_table = sorted(
            [
                ("|".join(strat_ids), " ".join(method.__doc__.split()))
                for method, strat_ids in method_to_ids.items()
            ]
        )

        # Reuse click machinery to format CLI helper output.
        formatter = ctx.make_formatter()
        # XXX Seems to have no effect. Should have introduced an empty line
        # before section.
        formatter.write_paragraph()
        with formatter.section("Available strategies"):
            formatter.write_dl(strat_table)

        click.echo(colorized_help(formatter.getvalue().rstrip(), keywords))
        ctx.exit()

    # Validate exclusive options requirement depending on strategy or action.
    validation_requirements = {
        strategy: (
            (
                regexp,
                "-r/--regexp",
                {
                    DISCARD_MATCHING_PATH,
                    DISCARD_NON_MATCHING_PATH,
                    SELECT_MATCHING_PATH,
                    SELECT_NON_MATCHING_PATH,
                },
            ),
        ),
        action: (
            (
                export,
                "-E/--export",
                {
                    COPY_SELECTED,
                    COPY_DISCARDED,
                    MOVE_SELECTED,
                    MOVE_DISCARDED,
                },
            ),
        ),
    }

    for conf_value, requirements in validation_requirements.items():
        for param_value, param_name, required_values in requirements:
            if conf_value in required_values:
                if not param_value:
                    raise click.BadParameter(
                        f"{conf_value} requires the {param_name} parameter."
                    )
            elif param_value:
                raise click.BadParameter(
                    f"{param_name} parameter not allowed in {conf_value}."
                )

    conf = Config(
        dry_run=dry_run,
        input_format=input_format,
        force_unlock=force_unlock,
        hash_only=hash_only,
        hash_headers=hash_header,
        size_threshold=size_threshold,
        content_threshold=content_threshold,
        show_diff=show_diff,
        strategy=strategy,
        time_source=time_source,
        regexp=regexp,
        action=action,
        export=export,
        export_format=export_format,
    )

    dedup = Deduplicate(conf)

    click.echo(title_style("\n● Phase #0 - Load mails"))
    with click.progressbar(
        mail_sources,
        length=len(mail_sources),
        label="Mail sources",
        show_pos=True,
    ) as progress:
        for source in progress:
            dedup.add_source(source)

    click.echo(title_style("\n● Phase #1 - Compute hashes and group duplicates"))
    dedup.hash_all()
    if hash_only:
        for all_mails in dedup.mails.values():
            for mail in all_mails:
                click.echo(mail.pretty_headers)
                click.echo(f"Hash: {mail.hash_key}")
        ctx.exit()

    click.echo(title_style("\n● Phase #2 - Select mails in each group"))
    dedup.select_all()

    click.echo(title_style("\n● Phase #3 - Perform action on selected mails"))
    perform_action(dedup)
    dedup.close_all()

    click.echo(title_style("\n● Phase #4 - Report and statistics"))
    # Print deduplication statistics, then performs a self-check on them.
    click.echo(dedup.report())
    dedup.check_stats()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mail_deduplicate', 'mail_deduplicate.tests']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=0.17.0,<0.18.0',
 'boltons>=20.2.1,<21.0.0',
 'click-help-colors>=0.8,<0.10',
 'click-log>=0.3.2,<0.4.0',
 'click>=7.1.2,<8.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'tomlkit>=0.7.0,<0.8.0']

extras_require = \
{'docs': ['sphinx>=3.4.2,<4.0.0', 'sphinx_rtd_theme>=0.5.1,<0.6.0']}

entry_points = \
{'console_scripts': ['mdedup = mail_deduplicate.cli:mdedup']}

setup_kwargs = {
    'name': 'mail-deduplicate',
    'version': '6.1.1',
    'description': '📧 CLI to deduplicate mails from mail boxes.',
    'long_description': 'Mail Deduplicate\n================\n\nCommand-line tool to deduplicate mails from a set of boxes.\n\nStable release: |release| |versions|\n\nDevelopment: |build| |docs| |coverage|\n\n.. |release| image:: https://img.shields.io/pypi/v/mail-deduplicate.svg\n    :target: https://pypi.python.org/pypi/mail-deduplicate\n    :alt: Last release\n.. |versions| image:: https://img.shields.io/pypi/pyversions/mail-deduplicate.svg\n    :target: https://pypi.python.org/pypi/mail-deduplicate\n    :alt: Python versions\n.. |build| image:: https://github.com/kdeldycke/mail-deduplicate/workflows/Tests/badge.svg\n    :target: https://github.com/kdeldycke/mail-deduplicate/actions?query=workflow%3ATests\n    :alt: Unittests status\n.. |docs| image:: https://readthedocs.org/projects/mail-deduplicate/badge/?version=develop\n    :target: https://mail-deduplicate.readthedocs.io/en/develop/\n    :alt: Documentation Status\n.. |coverage| image:: https://codecov.io/gh/kdeldycke/mail-deduplicate/branch/develop/graph/badge.svg\n    :target: https://codecov.io/github/kdeldycke/mail-deduplicate?branch=develop\n    :alt: Coverage Status\n\n.. figure:: https://raw.githubusercontent.com/kdeldycke/mail-deduplicate/develop/docs/cli-coloured-header.png\n    :align: center\n\n\nFeatures\n--------\n\n* Duplicate detection based on cherry-picked and normalized mail headers.\n* Source and deduplicate mails from multiple sources.\n* Reads and writes to ``mbox``, ``maildir``, ``babyl``, ``mh`` and ``mmdf`` formats.\n* Multiple duplicate selection strategies based on size, content, timestamp, file\n  path or random choice.\n* Copy, move or delete the resulting set of mails after the deduplication.\n* Dry-run mode.\n* Protection against false-positives by checking for size and content\n  differences.\n\n\nScreenshots\n-----------\n\n.. figure:: https://raw.githubusercontent.com/kdeldycke/mail-deduplicate/develop/docs/cli-colored-help.png\n    :align: center\n\n.. figure:: https://raw.githubusercontent.com/kdeldycke/mail-deduplicate/develop/docs/cli-coloured-run.png\n    :align: center\n\n\nInstallation\n------------\n\nThis package is `available on PyPi\n<https://pypi.python.org/pypi/mail-deduplicate>`_, so you can install the\nlatest stable release and its dependencies with a simple ``pip`` call:\n\n.. code-block:: shell-session\n\n    $ pip install mail-deduplicate\n\n\nDocumentation\n-------------\n\nDocs are `hosted on Read the Docs\n<https://mail-deduplicate.readthedocs.io>`_.\n',
    'author': 'Kevin Deldycke',
    'author_email': 'kevin@deldycke.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kdeldycke/mail-deduplicate',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import random
import string
import sys
from email.utils import formatdate as maildate
from mailbox import Mailbox, Maildir, Message, mbox
from textwrap import dedent, indent
from uuid import uuid4

import arrow
import click
import pytest
from boltons.iterutils import flatten, same
from boltons.tbutils import ExceptionInfo
from click.testing import CliRunner

from .. import CLI_NAME
from ..cli import mdedup

""" Fixtures, configuration and helpers for tests. """


def is_windows():
    """ Return `True` only if current platform is of the Windows family. """
    return sys.platform in ["win32", "cygwin"]


skip_windows = pytest.mark.skipif(is_windows(), reason="Skip Windows")
""" Pytest mark to skip a test if it is run on a Windows system. """


def print_cli_output(cmd, output):
    """ Simulate CLI output. Used to print debug traces in test results. """
    print("\n► {}".format(click.style(" ".join(cmd), fg="white")))
    if output:
        print(indent(output, "  "))


@pytest.fixture
def runner():
    runner = CliRunner()
    with runner.isolated_filesystem():
        yield runner


@pytest.fixture
def invoke(runner):
    """ Executes Click's CLI, print output and return results. """

    def _run(*args, color=False):
        # We allow for nested iterables and None values as args for
        # convenience. We just need to flatten and filters them out.
        args = list(filter(None.__ne__, flatten(args)))
        if args:
            assert same(map(type, args), str)

        result = runner.invoke(mdedup, args, color=color)

        print_cli_output([CLI_NAME] + args, result.output)

        # Print some more debug info.
        print(result)
        if result.exception:
            print(ExceptionInfo.from_exc_info(*result.exc_info).get_formatted())

        return result

    return _run


class MailFactory:

    """Create fake mail messages to serve as unittest fixtures.

    Help production of either random, customized or deterministic mail message.
    """

    def __init__(self, **custom_fields):
        """Init the mail with custom fields.

        You can bypass data normalization by passing the pre-formated date
        string with ``date_rfc2822`` custom field instead of ``date``.
        """
        # Defaults fields values.
        self.fields = {
            "body": "Да, они летят.",
            "date": arrow.utcnow(),
            "date_rfc2822": None,
            "message_id": "<201111231111.abcdef101@mail.nohost.com>",
        }

        # Check all custom fields are recognized and supported.
        assert set(custom_fields).issubset(self.fields)

        # Parse dates and normalize to Arrow instance.
        if "date" in custom_fields:
            custom_fields["date"] = arrow.get(custom_fields["date"])

        # Update default values with custom ones.
        self.fields.update(custom_fields)

        # Derive RFC-2822 date from arrow object if not set.
        if not self.fields.get("date_rfc2822"):
            self.fields["date_rfc2822"] = maildate(self.fields["date"].float_timestamp)

    def render(self):
        """ Returns the full, rendered content of the mail. """
        return dedent(
            """\
            Return-path: <none@nohost.com>
            Envelope-to: me@host.com
            Delivery-date: {date_rfc2822}
            Received: from [11.11.11.11] (helo=nope.com)
            \tby host.com with esmtp (Exim 4.80)
            \t(envelope-from <noone@nohost.com>)
            \tid 1CX8OJ-0014c9-Ii
            \tfor me@host.com; {date_rfc2822}
            Date: {date_rfc2822}
            From: foo@bar.com
            Message-Id: {message_id}
            To: baz
            Subject: A duplicate mail
            Mime-Version: 1.0
            Content-Length: 60
            Content-Type: text/plain; charset="utf-8"
            Content-Transfer-Encoding: 8bit
            {body}""".format(
                **self.fields
            )
        ).encode("utf-8")

    def as_message(self):
        """Returns the mail as an instance of ``mailbox.Message``."""
        return Message(self.render())

    @staticmethod
    def random_string(length):
        return "".join(random.choice(string.ascii_lowercase) for i in range(length))


@pytest.fixture
def make_box(tmp_path):
    """A generic fixture to produce a temporary box of mails.

    The mail container can be created in any format supported by Python standard
    library, by the way of the ``box_type`` parameter. Supported values: only
    ``Maildir`` and ``mbox`` for the moment.
    """

    def _make_box(box_type, mails=None):
        """Create a fake maildir and populate it with mails."""
        # Check parameters.
        assert box_type in (Maildir, mbox)
        assert issubclass(box_type, Mailbox)

        if not mails:
            mails = []
        assert same(map(type, mails), MailFactory)

        # Create the container under a random name and put all provided mails there.
        box = box_type(tmp_path.joinpath(uuid4().hex), create=True)
        box.lock()
        for fake_mail in mails:
            box.add(fake_mail.render())

        box.close()
        return box._path, box_type

    return _make_box


def check_box(box_path, box_type, content=None):
    """Check the content of a mail box (in any of maildir of mbox format).

    Does not use ``set()`` types internally to avoid silent deduplication.
    Translates all mails provided to ``mailbox.Message`` instances to provide
    fair comparison in a normalized space.
    """
    # Check provided parameters.
    assert isinstance(box_path, str)
    assert box_type in (Maildir, mbox)
    assert not isinstance(content, set)
    if content is None:
        content = []
    assert same(map(type, content), MailFactory)

    # Compares the content of the box.
    box = box_type(box_path, create=False)

    # TODO: use a COunter to count occurrences

    assert len(box) == len(content)
    mails_found = sorted([str(m) for m in box])
    assert sorted([str(m.as_message()) for m in content]) == mails_found
    box.close()



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from mailbox import Maildir
from string import ascii_lowercase

import arrow
import pytest

from ..strategy import (
    STRATEGY_METHODS,
    DISCARD_OLDER,
    DISCARD_OLDEST,
    DISCARD_NEWER,
    DISCARD_NEWEST,
    SELECT_OLDER,
    SELECT_OLDEST,
    SELECT_NEWER,
    SELECT_NEWEST,
    DISCARD_SMALLER,
    DISCARD_SMALLEST,
    DISCARD_BIGGER,
    DISCARD_BIGGEST,
    SELECT_SMALLER,
    SELECT_SMALLEST,
    SELECT_BIGGER,
    SELECT_BIGGEST,
    DISCARD_MATCHING_PATH,
    DISCARD_NON_MATCHING_PATH,
    SELECT_MATCHING_PATH,
    SELECT_NON_MATCHING_PATH,
    DISCARD_ONE,
    DISCARD_ALL_BUT_ONE,
    SELECT_ONE,
    SELECT_ALL_BUT_ONE,
)
from .conftest import MailFactory, check_box


def test_strategy_definitions():
    """ Test deduplication strategy definitions. """
    for strategy_id, method in STRATEGY_METHODS.items():
        # All strategies are lower cases strings, with dashes.
        assert isinstance(strategy_id, str)
        assert set(strategy_id).issubset(ascii_lowercase + "-")
        assert callable(method)


# Collections of pre-defined fixtures to use in the deduplication tests below.
smallest_mail = MailFactory(body="Hello I am a duplicate mail. With annoying ćĥäŖş.")
smaller_mail = MailFactory(body="Hello I am a duplicate mail. With annoying ćĥäŖş. ++")
bigger_mail = MailFactory(
    body="Hello I am a duplicate mail. With annoying ćĥäŖş. +++++"
)
biggest_mail = MailFactory(
    body="Hello I am a duplicate mail. With annoying ćĥäŖş. +++++++++"
)


# List of strategies and their required dummy options.
strategy_options = dict.fromkeys(STRATEGY_METHODS, [])
# Add dummy regexps.
strategy_options.update(
    {
        "discard-matching-path": ["--regexp=.*"],
        "discard-non-matching-path": ["--regexp=.*"],
        "select-matching-path": ["--regexp=.*"],
        "select-non-matching-path": ["--regexp=.*"],
    }
)


@pytest.mark.parametrize("strategy_id,options", strategy_options.items())
def test_maildir_dry_run(invoke, make_box, strategy_id, options):
    """ Check no mail is removed in dry-run mode. """
    box_path, box_type = make_box(
        Maildir,
        [
            smallest_mail,
            bigger_mail,
            smallest_mail,
            smaller_mail,
            smaller_mail,
            bigger_mail,
        ],
    )

    result = invoke(
        f"--strategy={strategy_id}",
        *options,
        "--action=delete-selected",
        "--dry-run",
        box_path,
    )

    assert result.exit_code == 0
    check_box(
        box_path,
        box_type,
        content=[
            smallest_mail,
            bigger_mail,
            smallest_mail,
            smaller_mail,
            smaller_mail,
            bigger_mail,
        ],
    )


@pytest.mark.parametrize("strategy_id", [SELECT_SMALLER, DISCARD_BIGGEST])
def test_maildir_smaller_strategy(invoke, make_box, strategy_id):
    """ Test strategy of small mail selection. """
    box_path, box_type = make_box(
        Maildir,
        [
            smallest_mail,
            biggest_mail,
            smallest_mail,
            bigger_mail,
            smaller_mail,
            smaller_mail,
            bigger_mail,
            biggest_mail,
        ],
    )

    result = invoke(f"--strategy={strategy_id}", "--action=delete-selected", box_path)

    assert result.exit_code == 0
    # Biggest mails are selected but not the smaller ones.
    check_box(
        box_path,
        box_type,
        content=[biggest_mail, biggest_mail],
    )


@pytest.mark.parametrize("strategy_id", [SELECT_SMALLEST, DISCARD_BIGGER])
def test_maildir_smallest_strategy(invoke, make_box, strategy_id):
    """ Test strategy of smallest mail selection. """
    box_path, box_type = make_box(
        Maildir,
        [
            smallest_mail,
            biggest_mail,
            smallest_mail,
            bigger_mail,
            smaller_mail,
            smaller_mail,
            bigger_mail,
            biggest_mail,
        ],
    )

    result = invoke(f"--strategy={strategy_id}", "--action=delete-selected", box_path)

    assert result.exit_code == 0
    # Bigger mails are selected but not the smallest ones.
    check_box(
        box_path,
        box_type,
        content=[
            biggest_mail,
            bigger_mail,
            smaller_mail,
            smaller_mail,
            bigger_mail,
            biggest_mail,
        ],
    )


@pytest.mark.parametrize("strategy_id", [SELECT_BIGGER, DISCARD_SMALLEST])
def test_maildir_bigger_strategy(invoke, make_box, strategy_id):
    """ Test strategy of bigger mail selection. """
    box_path, box_type = make_box(
        Maildir,
        [
            smallest_mail,
            biggest_mail,
            smallest_mail,
            bigger_mail,
            smaller_mail,
            smaller_mail,
            bigger_mail,
            biggest_mail,
        ],
    )

    result = invoke(f"--strategy={strategy_id}", "--action=delete-selected", box_path)

    assert result.exit_code == 0
    # Smallest mails are selected but not the bigger ones.
    check_box(
        box_path,
        box_type,
        content=[smallest_mail, smallest_mail],
    )


@pytest.mark.parametrize("strategy_id", [SELECT_BIGGEST, DISCARD_SMALLER])
def test_maildir_biggest_strategy(invoke, make_box, strategy_id):
    """ Test strategy of biggest mail selection. """
    box_path, box_type = make_box(
        Maildir,
        [
            smallest_mail,
            biggest_mail,
            smallest_mail,
            bigger_mail,
            smaller_mail,
            smaller_mail,
            bigger_mail,
            biggest_mail,
        ],
    )

    result = invoke(f"--strategy={strategy_id}", "--action=delete-selected", box_path)

    assert result.exit_code == 0
    # Smaller mails are selected but not the biggest ones.
    check_box(
        box_path,
        box_type,
        content=[
            smallest_mail,
            smallest_mail,
            bigger_mail,
            smaller_mail,
            smaller_mail,
            bigger_mail,
        ],
    )


newest_date = arrow.utcnow()
newer_date = newest_date.shift(minutes=-1)
older_date = newest_date.shift(minutes=-2)
oldest_date = newest_date.shift(minutes=-3)


newest_mail = MailFactory(date=newest_date)
newer_mail = MailFactory(date=newer_date)
older_mail = MailFactory(date=older_date)
oldest_mail = MailFactory(date=oldest_date)


@pytest.mark.parametrize("strategy_id", [SELECT_OLDER, DISCARD_NEWEST])
def test_maildir_older_strategy(invoke, make_box, strategy_id):
    """ Test strategy of older mail selection. """
    box_path, box_type = make_box(
        Maildir,
        [
            oldest_mail,
            newest_mail,
            oldest_mail,
            newer_mail,
            older_mail,
            older_mail,
            newer_mail,
            newest_mail,
        ],
    )

    result = invoke(f"--strategy={strategy_id}", "--action=delete-selected", box_path)

    assert result.exit_code == 0
    # Newest mails are selected but not the older ones.
    check_box(
        box_path,
        box_type,
        content=[newest_mail, newest_mail],
    )


@pytest.mark.parametrize("strategy_id", [SELECT_OLDEST, DISCARD_NEWER])
def test_maildir_oldest_strategy(invoke, make_box, strategy_id):
    """ Test strategy of oldest mail selection. """
    box_path, box_type = make_box(
        Maildir,
        [
            oldest_mail,
            newest_mail,
            oldest_mail,
            newer_mail,
            older_mail,
            older_mail,
            newer_mail,
            newest_mail,
        ],
    )

    result = invoke(f"--strategy={strategy_id}", "--action=delete-selected", box_path)

    assert result.exit_code == 0
    # Newer mails are selected but not the oldest ones.
    check_box(
        box_path,
        box_type,
        content=[
            newest_mail,
            newer_mail,
            older_mail,
            older_mail,
            newer_mail,
            newest_mail,
        ],
    )


@pytest.mark.parametrize("strategy_id", [SELECT_NEWER, DISCARD_OLDEST])
def test_maildir_newer_strategy(invoke, make_box, strategy_id):
    """ Test strategy of newer mail selection. """
    box_path, box_type = make_box(
        Maildir,
        [
            oldest_mail,
            newest_mail,
            oldest_mail,
            newer_mail,
            older_mail,
            older_mail,
            newer_mail,
            newest_mail,
        ],
    )

    result = invoke(f"--strategy={strategy_id}", "--action=delete-selected", box_path)

    assert result.exit_code == 0
    # Oldest mails are selected but not the newer ones.
    check_box(
        box_path,
        box_type,
        content=[oldest_mail, oldest_mail],
    )


@pytest.mark.parametrize("strategy_id", [SELECT_NEWEST, DISCARD_OLDER])
def test_maildir_newest_strategy(invoke, make_box, strategy_id):
    """ Test strategy of newest mail selection. """
    box_path, box_type = make_box(
        Maildir,
        [
            oldest_mail,
            newest_mail,
            oldest_mail,
            newer_mail,
            older_mail,
            older_mail,
            newer_mail,
            newest_mail,
        ],
    )

    result = invoke(f"--strategy={strategy_id}", "--action=delete-selected", box_path)

    assert result.exit_code == 0
    # Older mails are selected but not the newest ones.
    check_box(
        box_path,
        box_type,
        content=[
            oldest_mail,
            oldest_mail,
            newer_mail,
            older_mail,
            older_mail,
            newer_mail,
        ],
    )


random_mail_1 = MailFactory(message_id=MailFactory.random_string(30))
random_mail_2 = MailFactory(message_id=MailFactory.random_string(30))


@pytest.mark.parametrize("strategy_id", [SELECT_ONE, DISCARD_ALL_BUT_ONE])
def test_maildir_one_strategy(invoke, make_box, strategy_id):
    """ Test strategy of discarding one random duplicate. """
    box_path, box_type = make_box(
        Maildir,
        [
            random_mail_1,
            random_mail_1,
            random_mail_1,
            random_mail_2,
            random_mail_2,
            random_mail_2,
        ],
    )

    result = invoke(f"--strategy={strategy_id}", "--action=delete-selected", box_path)

    assert result.exit_code == 0
    # Newest mails are selected but not the older ones.
    check_box(
        box_path,
        box_type,
        content=[
            random_mail_1,
            random_mail_1,
            random_mail_2,
            random_mail_2,
        ],
    )


@pytest.mark.parametrize("strategy_id", [SELECT_ALL_BUT_ONE, DISCARD_ONE])
def test_maildir_all_but_one_strategy(invoke, make_box, strategy_id):
    """ Test strategy of discarding all but one random duplicate. """
    box_path, box_type = make_box(
        Maildir,
        [
            random_mail_1,
            random_mail_1,
            random_mail_1,
            random_mail_2,
            random_mail_2,
            random_mail_2,
        ],
    )

    result = invoke(f"--strategy={strategy_id}", "--action=delete-selected", box_path)

    assert result.exit_code == 0
    # Newest mails are selected but not the older ones.
    check_box(
        box_path,
        box_type,
        content=[
            random_mail_1,
            random_mail_2,
        ],
    )
